import math


def c_mod(a: int, n: int) -> int:
    return a % n if a >= 0 else a % n - n


def NormalizeAngle(Angle: float, FullCircle: int) -> float:
    if (Angle >= 0) and (Angle < FullCircle):
        ang = Angle
    else:
        TruncAng = math.trunc(Angle)
        ang = (TruncAng % FullCircle) + Angle - TruncAng
        # mod function is different between python and C. thus the following C code is redundant in python
        # if Angle < 0:
        #     ang = FullCircle + ang
    return ang


def NormalizeAngleDeg(Angle: float) -> float:
    return NormalizeAngle(Angle, 360)



from czml3 import Document, Packet, Preamble
from czml3.properties import (
    Material,
    RectangleCoordinates,
    Rectangle,
    ImageMaterial,
)

from osgeo import gdal
from gdalos import gdalos_trans, projdef
import base64


def gdal_to_czml(ds, name=None, out_filename=None, description=None):
    pjstr_src_srs = projdef.get_srs_pj_from_ds(ds)
    pjstr_tgt_srs = projdef.get_srs_pj_from_epsg()
    if not projdef.proj_is_equivalent(pjstr_src_srs, pjstr_tgt_srs):
        ds = gdalos_trans(ds, warp_CRS=pjstr_tgt_srs, of='MEM', ovr_type=None, write_spec=False)

    # calculate the extent
    ulx, xres, xskew, uly, yskew, yres = ds.GetGeoTransform()
    lrx = ulx + (ds.RasterXSize * xres)
    lry = uly + (ds.RasterYSize * yres)
    wsen = [ulx, lry, lrx, uly]

    # http://osgeo-org.1560.x6.nabble.com/GDAL-Python-Save-a-dataset-to-an-in-memory-Python-Bytes-object-td5280254.html
    # reading the gdal raster data into a PNG memory buffer
    gdal.GetDriverByName('PNG').CreateCopy('/vsimem/output.png', ds)

    # Read /vsimem/output.png
    f = gdal.VSIFOpenL('/vsimem/output.png', 'rb')
    gdal.VSIFSeekL(f, 0, 2)  # seek to end
    size = gdal.VSIFTellL(f)
    gdal.VSIFSeekL(f, 0, 0)  # seek to beginning
    png_data = gdal.VSIFReadL(1, size, f)
    gdal.VSIFCloseL(f)

    # Cleanup
    gdal.Unlink('/vsimem/output.png')

    # encoding the png into base64
    base64_data = base64.b64encode(png_data)

    czml_doc = Document(
        [
            Preamble(
                name="czml",
                description=description,
            ),
            Packet(
                id="rect",
                name=name,
                rectangle=Rectangle(
                    coordinates=RectangleCoordinates(wsenDegrees=wsen),
                    fill=True,
                    material=Material(
                        image=ImageMaterial(
                            transparent=True,
                            repeat=None,
                            image=
                                "data:image/png;base64,"+base64_data.decode("utf-8")
                        ),
                    ),
                ),
            ),
        ]
    )
    ds = None  # close the ds in case it was transformed
    if out_filename:
        with open(out_filename, 'w') as f:
            print(czml_doc, file=f)
    return czml_doc



import os
import tempfile
from typing import Optional, Sequence, List, Union
from osgeo import gdal, ogr, osr
from gdalos.rectangle import GeoRectangle
from gdalos.calc import gdal_to_czml
from gdalos.gdalos_color import ColorPalette, get_file_from_strings
from gdalos import gdalos_util

# def get_named_temporary_filenme(suffix='', dir_name=''):
#     filename = next(tempfile._get_candidate_names())+suffix
#     if dir_name is None:
#         return filename
#     else:
#         if dir_name=='':
#             dir_name = tempfile._get_default_tempdir()
#         return os.path.join(dir_name, filename)


def gdal_crop(ds: gdal.Dataset, out_filename: str, output_format: str = 'MEM',
              extent: Optional[GeoRectangle] = None,
              cutline: Optional[Union[str, List[str]]] = None,
              common_options: dict = None):
    common_options = dict(common_options or dict())
    translate_options = {}
    warp_options = {}
    temp_filename = None

    if cutline is None:
        if extent is None:
            return ds
        # -projwin minx maxy maxx miny (ulx uly lrx lry)
        translate_options['projWin'] = extent.lurd
        # -te minx miny maxx maxy
    else:
        if extent is not None:
            warp_options['outputBounds'] = extent.ldru
        warp_options['cropToCutline'] = extent is None
        warp_options['dstNodata'] = -32768
        # warp_options['dstAlpha'] = True

        if isinstance(cutline, str):
            cutline_filename = cutline
        elif isinstance(cutline, Sequence):
            temp_filename = tempfile.mktemp(suffix='.gpkg')
            cutline_filename = temp_filename
            gdalos_util.wkt_write_ogr(cutline_filename, cutline, of='GPKG')
        warp_options['cutlineDSName'] = cutline_filename
    # else:
    #     raise Exception("extent is unknown {}".format(extent))

    common_options['format'] = output_format
    if warp_options:
        ds = gdal.Warp(str(out_filename), ds, **common_options, **warp_options)
    else:
        ds = gdal.Translate(str(out_filename), ds, **common_options, **translate_options)
    if temp_filename is not None:
        os.remove(temp_filename)
    return ds


def make_czml_description(pal:ColorPalette, process_palette):
    if pal:
        if process_palette >= 2:
            return ' '.join(['{}:{}'.format(
                ColorPalette.format_number(x),
                ColorPalette.format_color(c)) for x, c in pal.pal.items()])
        else:
            return ' '.join([ColorPalette.format_number(x) for x in pal.pal.keys()])
    else:
        return None


def czml_gdaldem_crop_and_color(ds: gdal.Dataset, process_palette, czml_output_filename: str, **kwargs):
    ds, pal = gdaldem_crop_and_color(ds=ds, process_palette=process_palette, **kwargs)
    if ds is None:
        raise Exception('fail to color')
    if czml_output_filename is not None:
        description = make_czml_description(pal, process_palette)
        gdal_to_czml.gdal_to_czml(ds, name=czml_output_filename, out_filename=czml_output_filename, description=description)
    return ds


def gdaldem_crop_and_color(ds: gdal.Dataset,
                           out_filename: str, output_format: str = 'GTiff',
                           extent: Optional[GeoRectangle] = None,
                           cutline: Optional[Union[str, List[str]]] = None,
                           color_palette: Optional[Union[str, Sequence[str]]] = None,
                           process_palette=2,
                           common_options: dict = None):
    do_color = color_palette is not None
    do_crop = (extent or cutline) is not None

    if out_filename is None:
        out_filename = ''
        if output_format != 'MEM':
            raise Exception('output filename is None')

    if do_crop:
        output_format_crop = 'MEM' if do_color else output_format
        ds = gdal_crop(ds, out_filename,
                       output_format=output_format_crop, extent=extent, cutline=cutline,
                       common_options=common_options)
        if ds is None:
            raise Exception('fail to crop')

    pal = None
    if process_palette:
        bnd = ds.GetRasterBand(1)
        bnd.ComputeStatistics(0)
        min_val = bnd.GetMinimum()
        max_val = bnd.GetMaximum()

    if do_color:
        color_filename, temp_color_filename = get_file_from_strings(color_palette)
        if not process_palette:
            pal = None
        else:
            pal = ColorPalette()
            pal.read(color_filename)
            pal.apply_percent(min_val, max_val)
            # color_palette_stats(color_filename, min_val, max_val, process_palette)
        dem_options = {
            'addAlpha': True,
            'format': output_format,
            'processing': 'color-relief',
            'colorFilename': color_filename}
        ds = gdal.DEMProcessing(out_filename, ds, **dem_options)
        if temp_color_filename is not None:
            os.remove(temp_color_filename)
        if ds is None:
            raise Exception('fail to color')
    # else:
    #     stats = [min_val, max_val]
    return ds, pal


def get_wkt_list(filename):
    ds = ogr.Open(filename, 0)
    layer = ds.GetLayer()
    wkt_list = []
    for feature in layer:
        geom = feature.GetGeometryRef()
        wkt_list.append(geom.ExportToWkt())
    return wkt_list


def read_list(filename):
    return [line.rstrip('\n') for line in open(filename)]


if __name__ == '__main__':
    script_dir = os.path.dirname(os.path.realpath(__file__))
    root_data = os.path.join(script_dir, r'../../data/sample')
    shp_filename = os.path.join(root_data, r'shp/poly.shp')
    color_palette_filename = os.path.join(root_data, r'color_files/percents.txt')
    wkt_list = get_wkt_list(shp_filename)
    color_palette = read_list(color_palette_filename)
    raster_filename = os.path.join(root_data, r'maps/srtm1_x35_y32.tif')
    ds = gdalos_util.open_ds(raster_filename)
    out_filename = tempfile.mktemp(suffix='.tif')
    ds, pal = gdaldem_crop_and_color(
        ds=ds,
        out_filename=out_filename,
        cutline=wkt_list,
        color_palette=color_palette,
        output_format='GTiff')
    print(out_filename)
    print(pal)



import re

from fidget.backend.QtWidgets import QHBoxLayout
from fidget.core import PlaintextParseError, inner_plaintext_parser, validator
from fidget.widgets import (
    FidgetCombo,
    FidgetConverter,
    FidgetDict,
    FidgetEditCombo,
    FidgetFloat,
    FidgetLine,
    FidgetStacked,
    inner_fidget,
)


@validator("zone must be between 1 and 60")
def zone_validator(v: float):
    return 1 <= v <= 60


class CrsWidget(FidgetStacked[str]):
    @inner_fidget()
    class CrsWidgetUtm(FidgetConverter[dict, str]):
        @inner_fidget("utm", make_indicator=False, make_title=False)
        class _CrsWidgetUtmMulti(FidgetDict):
            INNER_TEMPLATES = [
                FidgetEditCombo.template(
                    "datum",
                    ("w84", "e50"),
                    make_title=False,
                    make_indicator=False,
                    make_plaintext=False,
                ),
                FidgetFloat.template(
                    "zone",
                    make_title=False,
                    validation_func=zone_validator,
                    make_indicator=False,
                ),
            ]

            LAYOUT_CLS = QHBoxLayout

        def convert(self, d: dict):
            return f'{d["datum"]}u{d["zone"]:n}'

        @inner_plaintext_parser
        @staticmethod
        def shorthand(s: str):
            split = s.rsplit("u", 2)
            if len(split) == 1:
                raise PlaintextParseError('string must contain a "u" separator')
            return s

        def back_convert(self, s: str):
            split = s.rsplit("u", 2)
            if len(split) == 1:
                datum, zone = s, ""
            else:
                datum, zone = split

            try:
                zone = float(zone)
            except ValueError:
                zone = 0

            return {"datum": datum, "zone": zone}

    @inner_fidget("proj4")
    class CrsWidgetProj(FidgetLine):
        PATTERN = re.compile(
            r"\+proj=[a-zA-Z0-9_.]+(\s+\+[a-zA-Z_][a-zA-Z0-9_]*=[-a-zA-Z0-9_.]+)*"
        )

    @inner_fidget(
        "builtin", options=("w84geo",), initial_value="w84geo", make_title=False
    )
    class CrsWidgetBuiltin(FidgetCombo):
        pass

    MAKE_TITLE = True
    MAKE_INDICATOR = True
    LAYOUT_CLS = QHBoxLayout
    SELECTOR_CLS = "radio"


if __name__ == "__main__":
    from fidget.backend.QtWidgets import QApplication
    from fidget.widgets import FidgetOptional

    app = QApplication([])
    w = FidgetOptional(CrsWidget("sample", make_plaintext=True))
    w.show()
    app.exec_()
    print(w.value())



# http://resources.esri.com/help/9.3/arcgisdesktop/com/gp_toolref/spatial_analyst_tools/how_viewshed_works.htm
# https://desktop.arcgis.com/en/arcmap/10.3/tools/spatial-analyst-toolbox/using-viewshed-and-observer-points-for-visibility.htm
# For applications involving visibility with radio waves,
# the refraction correction to be applied depends on the wavelength of the signal.
# The location of telecommunication sites is not merely a simple matter of determining the intervisibility
# but, rather, involves a number of parameters associated with the modeling of radio wave propagation
# including reflection, refraction specific to the frequency, attenuation (signal weakening),
# interference, atmospheric effects, and so on.
# Nonetheless, Observer Points and Viewshed are appropriate tools to use during the preliminary
# investigation stages of assessing possible telecommunication sites and coverage.

# http://web.soccerlab.polymtl.ca/grass/monsterViewer.php?functionName=main&fileName=/data/project-manager/grass/grassSVN/grass-addons/raster/r.viewshed/main.cc&iframe=true&width=100%&height=100%#cloneCode
# atmospheric refraction coeff. 1/7 for visual, 0.325 for radio waves,
# in future we might calculate this based on the physics, for now we just fudge by the 1/7th approximation.

# https://invest.readthedocs.io/en/3.4.1/final_ecosystem_services/scenic_quality.html
# refraction (float) – The earth curvature correction option corrects for the curvature of the
# earth and refraction of visible light in air. Changes in air density curve the light downward causing
# an observer to see further and the earth to appear less curved. While the magnitude of this effect varies
# with atmospheric conditions, a standard rule of thumb is that refraction of visible light reduces the
# apparent curvature of the earth by one-seventh. By default, this model corrects for the curvature of the
# earth and sets the refractivity coefficient to 0.13.

# https://gdal.org/programs/gdal_viewshed.html
# https://grass.osgeo.org/grass78/manuals/r.viewshed.html
# grass refraction_coeff  = 1/7 ~= 0.14286
# gdal: curve_coefficient(cc) = 1-refraction_coeff = 0.85714 ~= 6/7
# gdal: height_corrected = dem_height - cc * target_distance^2/sphere_diameter = dem_height
# talos: earth_curve_factor = 1/cc = 1/(1-refraction_coeff); refraction_coeff = 1-1/earth_curve_factor

# refraction_coeff values
# 0 -> normal sphere without correction (earth_curve_factor = 1)
# 1/7 ~= 0.14286 -> normal correction for visible light (earth_curve_factor = 7/6)
# 0.25~0.325 -> radio waves (earth_curve_factor = 1.33~1.48)
# 1 -> flat earth (earth_curve_factor = inf)

# earth_curv_factor = (1-refraction_coeff) / SphereDiameter;


def height_correction(target_distance, refraction_coeff, sphere_radius=6378137):
    sphere_diameter = sphere_radius*2
    cc = 1-refraction_coeff
    return - cc * target_distance**2/sphere_diameter


# from itertools import chain
# lst = list(chain.from_iterable(((10*10**i, 50*10**i) for i in range(1, 6))))
lst = (100, 500, 1_000, 5_000, 10_000, 50_000, 100_000, 500_000)
for d in lst:
    for rf in (0, 1/7, 1/4, 0.325):
        print('d: {} rf: {:.5f} height_correction: {:2f}'.format(d, rf, height_correction(d, rf)))
    print('')



from copy import copy
from gdalos.viewshed.viewshed_params import ViewshedParams


class ViewshedGridParams(ViewshedParams):
    __slots__ = ['name', 'interval', 'grid_range']

    def __init__(self):
        super().__init__()
        self.max_r = 2000
        self.ox = 700_000
        self.oy = 3550_000
        self.oz = 10
        self.tz = 10

        # self.azimuth = 20
        # self.h_aperture = 30

        self.name = None
        self.interval = self.max_r / 2
        j = 1
        self.grid_range = range(-j, j + 1)

    def get_array(self):
        result = []
        prefix = self.name + '_' if self.name else ''
        for i in self.grid_range:
            for j in self.grid_range:
                res = copy(self)
                res.ox = self.ox + i * self.interval
                res.oy = self.oy + j * self.interval
                res.name = prefix+'{}_{}'.format(i, j)
                result.append(res)
        return result

    def get_as_gdal_params_array(self):
        res = copy(self)
        res.ox = []
        res.oy = []
        for i in self.grid_range:
            for j in self.grid_range:
                ox = self.ox + i * self.interval
                oy = self.oy + j * self.interval
                res.ox.append(ox)
                res.oy.append(oy)
        return res.get_as_gdal_params()



from numbers import Real
from osgeo import gdal
from gdalos import gdalos_trans, GeoRectangle, gdalos_extent, gdalos_util, projdef
from pathlib import Path

# https://www.programcreek.com/python/example/101827/gdal.RasterizeLayer
# https://trac.osgeo.org/gdal/ticket/5581
# https://github.com/OSGeo/gdal/blob/master/gdal/apps/gdal_rasterize_lib.cpp


def gdalos_rasterize(
        in_filename: str, shp_filename_or_ds: str,
        out_filename: str = None, shp_layer_name: str = None, shp_z_attribute: str = 'Height',
        add: bool = True, extent: GeoRectangle = ..., out_res: Real = None, out_srs: str = None, overwrite: bool = True):

    """rasterize a vector layer into a given raster layer, overriding or adding the values

    Parameters
    ----------
    in_filename:str
        input raster filename
    out_filename:str
        ouput raster filename, if None - input raster would be updated inplace
    shp_filename_or_ds: str
        input vector filename or dataset
    shp_layer_name: str
        name of the layer from the vector dataset to use
    shp_z_attribute: str='Height'
        name of the attribute to extract the z value from
    add: bool=True
        True to add to dtm values to shape values into the raster, False to burn shape values into the raster
    extent: GeoRectangle = ...
        None - use the extent of the input raster
        ... - use the extent of the input vector layer
        GeoRectangle - custom extent
    out_res: Real
        output resolution (if None then auto select)
    out_srs: str=None
        output srs
    overwrite: bool=True
        what to do if the output exists (fail of overwrite)

    Returns
    -------
        output raster dataset
    """

    # shp = gdalos_util.open_ds(shp_filename, gdal.gdalconst.OF_VECTOR)
    if shp_filename_or_ds is None:
        shp = None
    elif isinstance(shp_filename_or_ds, (str, Path)):
        shp = gdal.OpenEx(str(shp_filename_or_ds), gdal.gdalconst.OF_VECTOR)
    else:
        shp = shp_filename_or_ds

    if out_filename is None:
        dstDs = gdalos_util.open_ds(in_filename, gdal.GA_Update)
    else:
        pj4326 = projdef.get_srs_pj_from_epsg()
        if extent is ...:
            cov_pj_srs = projdef.get_srs_pj_from_ds(shp)
            if not cov_pj_srs:
                cov_pj_srs = pj4326
            cov_extent = gdalos_extent.get_vec_extent(shp.GetLayer())
            transform = projdef.get_transform(cov_pj_srs, pj4326)
            extent = gdalos_extent.translate_extent(cov_extent, transform)

        dstDs = gdalos_trans(in_filename, out_filename, extent=extent, warp_CRS=out_srs, out_res=out_res,
                             cog=False, ovr_type=None, overwrite=overwrite, return_ds=True)

    if shp is not None:
        rasteize_options = gdal.RasterizeOptions(
            layers=shp_layer_name, add=add, attribute=shp_z_attribute)
        ret = gdal.Rasterize(dstDs, shp, options=rasteize_options)
        if ret != 1:
            raise Exception('Rasterize failed')

    return dstDs


if __name__ == '__main__':
    out_dir = Path(r'd:\temp\rasterize')
    my_out_filename = out_dir / 'output.tif'
    cov_filename = out_dir / Path('ras_cov.gpkg')

    my_out_srs = projdef.get_proj_string(36)
    do_overwrite = True
    do_cog = True
    do_add = False
    do_shp = True
    do_inplace = False

    if do_shp:
        shp_filename_or_ds = cov_filename
        out_extent = ...
    else:
        shp_filename_or_ds = None
        out_extent = GeoRectangle.from_min_max(34.0, 34.2, 32.0, 32.2)

    if do_inplace:
        input_raster = out_dir / 'input.tif'
        out_filename = None
    else:
        input_raster = Path(r'd:\Maps\w84geo\dtm\SRTM1_hgt.tif.x20-80_y20-40.cog.tif')
        out_filename = my_out_filename

    dstDs = gdalos_rasterize(
        input_raster,
        shp_filename_or_ds=shp_filename_or_ds, out_filename=out_filename,
        add=do_add, extent=out_extent, out_srs=my_out_srs, overwrite=do_overwrite)

    if do_cog:
        if out_filename is None:
            out_filename = my_out_filename
        gdalos_trans(dstDs, out_filename=str(out_filename) + '.cog.tif', overwrite=do_overwrite)

    del dstDs
    print('done!')





import os
import re
import gdal
from xml.dom import minidom
from collections import OrderedDict
from pathlib import Path
import glob
import tempfile
from typing import Sequence


def to_number(s):
    try:
        return int(s)
    except ValueError:
        return float(s)


def byte(number, i):
    return (number & (0xff << (i * 8))) >> (i * 8)


class ColorPalette:
    __slots__ = ['pal', '_all_numeric']

    def __init__(self):
        self.pal = OrderedDict()
        self._all_numeric = True

    def __repr__(self):
        return str(self.pal)

    def apply_percent(self, min_val, max_val):
        if self._all_numeric:
            # nothing to do
            return
        all_numeric = True
        new_pal = self.pal.copy()
        for val in self.pal.keys():
            if not isinstance(val, str):
                continue
            is_percent = val.endswith('%')
            if is_percent:
                new_val = val.rstrip('%')
                try:
                    new_val = to_number(new_val)
                    if is_percent:
                        new_val = (max_val - min_val) * new_val * 0.01 + min_val
                    new_pal[new_val] = new_pal.pop(val)
                except ValueError:
                    all_numeric = False
            else:
                all_numeric = False
                continue
        if all_numeric:
            self._all_numeric = True
        self.pal = new_pal

    def read(self, filename):
        if Path(filename).suffix.lower() == '.qlr':
            self.read_qlr(filename)
        else:
            self.read_color_file(filename)

    def read_color_file(self, color_filename):
        self.pal.clear()
        with open(str(color_filename)) as fp:
            for line in fp:
                split_line = line.strip().split(' ', maxsplit=1)
                if len(split_line) < 2:
                    continue
                try:
                    color = self.pal_color_to_rgb(split_line[1])
                    key = split_line[0].strip()
                except:
                    raise Exception('Error reading palette line: {}'.format(line))
                try:
                    key = to_number(key)
                except ValueError:
                    # should be percent
                    self._all_numeric = False
                    pass
                self.pal[key] = color

    def write_color_file(self, color_filename):
        with open(str(color_filename), mode='w') as fp:
            for key, color in self.pal.items():
                cc = self.color_to_cc(color)
                cc = ' '.join(str(c) for c in cc)
                fp.write('{} {}\n'.format(key, cc))

    def read_qlr(self, qlr_filename):
        self.pal.clear()
        qlr = minidom.parse(str(qlr_filename))
        #             <paletteEntry color="#ffffff" alpha="0" label="0" value="0"/>
        color_palette = qlr.getElementsByTagName("paletteEntry")
        for palette_entry in color_palette:
            color = palette_entry.getAttribute("color")
            if str(color).startswith('#'):
                color = int(color[1:], 16)
            alpha = int(palette_entry.getAttribute("alpha"))
            color = color + (alpha << 8*3)  # * 256**3
            key = int(palette_entry.getAttribute("value"))
            self.pal[key] = color

    def get_color_table(self, min_val=0, max_val=256, fill_missing_colors=True):
        # create color table
        color_table = gdal.ColorTable()
        if fill_missing_colors:
            keys = list(self.pal.keys())
            vals = list(self.pal.values())
            c = self.color_to_cc(vals[0])
            for key in range(min_val, max_val):
                if key in keys:
                    c = self.color_to_cc(self.pal[key])
                color_table.SetColorEntry(key, c)
        else:
            for key, col in self.pal.items():
                color_table.SetColorEntry(key, self.color_to_cc(col))  # set color for each key

        return color_table

    @staticmethod
    def format_number(num):
        return num if isinstance(num, str) else '{:.2f}'.format(num)

    @staticmethod
    def format_color(col):
        return col if isinstance(col, str) else '#{:06X}'.format(col)

    @staticmethod
    def color_to_cc(color):
        # if color < 256:
        #     return color
        # else:
            b = byte(color, 0)
            g = byte(color, 1)
            r = byte(color, 2)
            a = byte(color, 3)

            if a < 255:
                return r, g, b, a
            else:
                return r, g, b

    @staticmethod
    def pal_color_to_rgb(cc):
        # r g b a -> argb
        # todo: support color names or just find the gdal implementation of this function...
        # cc = color components
        cc = re.findall(r'\d+', cc)
        try:
            # if not rgb_colors:
            #     return (*(int(c) for c in cc),)
            if len(cc) == 1:
                return int(cc[0])
            elif len(cc) == 3:
                return (((((255 << 8) + int(cc[0])) << 8) + int(cc[1])) << 8) + int(cc[2])
            elif len(cc) == 4:
                return (((((int(cc[3]) << 8) + int(cc[0])) << 8) + int(cc[1])) << 8) + int(cc[2])
            else:
                return 0
        except:
            return 0


def qlr_to_color_file(qlr_filename: Path) -> Path:
    qlr_filename = Path(qlr_filename)
    cp = ColorPalette()
    cp.read_qlr(qlr_filename)
    color_filename = qlr_filename.with_suffix('.txt')
    cp.write_color_file(color_filename)
    return color_filename


def get_file_from_strings(color_palette):
    temp_color_filename = None
    if isinstance(color_palette, (Path, str)):
        color_filename = color_palette
    elif isinstance(color_palette, Sequence):
        temp_color_filename = tempfile.mktemp(suffix='.txt')
        color_filename = temp_color_filename
        with open(temp_color_filename, 'w') as f:
            for item in color_palette:
                f.write(item+'\n')
    else:
        raise Exception('Unknown color palette type {}'.format(color_palette))
    return color_filename, temp_color_filename


def get_color_table(color_palette):
    if color_palette is None:
        return None
    color_filename, temp_color_filename = get_file_from_strings(color_palette)
    pal = ColorPalette()
    pal.read(color_filename)
    color_table = pal.get_color_table()
    if temp_color_filename:
        os.remove(temp_color_filename)
    return color_table


if __name__ == "__main__":
    # dir_path = Path('/home/idan/maps/comb')
    dir_path = Path(r'sample/color_files')
    for filename in glob.glob(str(dir_path / '*.qlr')):
        qlr_to_color_file(filename)



from functools import wraps
from inspect import Parameter, signature
from itertools import chain
from typing import Mapping, Sequence


class CallParamDict:
    def __init__(
        self,
        func,
        args: tuple,
        kwargs: dict,
        pos_param_names: Sequence[str],
        all_params: Mapping[str, Parameter],
    ):
        self.func = func

        self.args = args
        self.kwargs = kwargs

        self.pos_param_names = pos_param_names
        self.all_params = all_params

        self.all_arguments = dict(zip(pos_param_names, args))
        self.all_arguments.update(kwargs)

    def __str__(self):
        return f"""{self.func.__name__}({
            ",".join(chain(
                (repr(a) for a in self.args),
                (f'{k}= {v!r}' for (k,v) in self.kwargs.items()))
            )
        })"""

    def __getitem__(self, item):
        if item in self.all_arguments:
            return self.all_arguments[item]

        p = self.all_params[item]
        if p.default != Parameter.empty:
            return p.default

        raise KeyError(item)

    def __setitem__(self, key, value):
        prev = key in self.all_arguments
        self.all_arguments[key] = value
        if prev:
            # value was default before, nothing more needs changing
            return


def with_param_dict(kwarg_name="_params"):
    def decorator(func):
        all_parameters: Mapping[str, Parameter] = signature(func).parameters
        dest_param = all_parameters.get(kwarg_name)
        if not dest_param or dest_param.kind != Parameter.KEYWORD_ONLY:
            raise NameError(
                "function must contain keyword-only parameter named " + kwarg_name
            )

        pos_names = []
        for n, p in all_parameters.items():
            if p.kind == Parameter.VAR_POSITIONAL:
                raise TypeError(
                    f"with_param_dict can't a variadic argument parameter ({p})"
                )
            if p.kind not in (
                Parameter.POSITIONAL_ONLY,
                Parameter.POSITIONAL_OR_KEYWORD,
            ):
                break
            pos_names.append(n)

        @wraps(func)
        def wrapper(*args, **kwargs):
            if kwarg_name in kwargs:
                raise TypeError(f"{kwarg_name} cannot be specified outside the wrapper")

            params = dict(zip(pos_names, args))
            params.update(kwargs)
            for k, p in all_parameters.items():
                if k == kwarg_name:
                    continue
                params.setdefault(k, p.default)

            kwargs[kwarg_name] = params

            return func(*args, **kwargs)

        return wrapper

    return decorator



import glob
import os
from pathlib import Path
from typing import Iterator, Sequence

import gdal, ogr, osr


def open_ds(filename_or_ds, *args, **kwargs):
    ods = OpenDS(filename_or_ds, *args, **kwargs)
    return ods.__enter__()


class OpenDS:
    def __init__(self, filename_or_ds, *args, **kwargs):
        if isinstance(filename_or_ds, (str, Path)):
            self.filename = str(filename_or_ds)
            self.ds = None
        else:
            self.filename = None
            self.ds = filename_or_ds
        self.args = args
        self.kwargs = kwargs
        self.own = None

    def __enter__(self) -> gdal.Dataset:
        if self.ds is None:
            self.ds = self._open_ds(self.filename, *self.args, **self.kwargs)
            if self.ds is None:
                raise IOError('could not open file "{}"'.format(self.filename))
            self.own = True
        return self.ds

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.own:
            self.ds = None

    @staticmethod
    def _open_ds(
            filename,
            access_mode=gdal.GA_ReadOnly,
            src_ovr: int = None,
            open_options: dict = None,
            logger=None,
    ):
        open_options = dict(open_options or dict())
        if src_ovr is not None and src_ovr >= 0:
            open_options["OVERVIEW_LEVEL"] = src_ovr
        if logger is not None:
            s = 'openning file: "{}"'.format(filename)
            if open_options:
                s = s + " with options: {}".format(str(open_options))
            logger.debug(s)
        if open_options:
            open_options = ["{}={}".format(k, v) for k, v in open_options.items()]

        if open_options:
            return gdal.OpenEx(str(filename), open_options=open_options)
        else:
            return gdal.Open(str(filename), access_mode)



def _get_bands(ds: gdal.Dataset) -> Iterator[gdal.Band]:
    return (ds.GetRasterBand(i + 1) for i in range(ds.RasterCount))


def _band_getmin(band):
    ret = band.GetMinimum()
    if ret is None:
        band.ComputeStatistics(0)
    return band.GetMinimum()


def get_band_types(filename_or_ds):
    with OpenDS(filename_or_ds) as ds:
        return [gdal.GetDataTypeName(band.DataType) for band in _get_bands(ds)]


def get_raster_band(filename_or_ds, bnd_index=1, ovr_index=None):
    with OpenDS(filename_or_ds) as ds:
        bnd = ds.GetRasterBand(bnd_index)
        if ovr_index is not None:
            bnd = bnd.GetOverview(ovr_index)
        return bnd


def get_ovr_count(filename_or_ds):
    with OpenDS(filename_or_ds) as ds:
        bnd = ds.GetRasterBand(1)
        return bnd.GetOverviewCount()


def get_nodatavalue(filename_or_ds):
    with OpenDS(filename_or_ds) as ds:
        band = next(_get_bands(ds))
        return band.GetNoDataValue()


def unset_nodatavalue(filename_or_ds):
    with OpenDS(filename_or_ds, access_mode=gdal.GA_Update) as ds:
        for b in _get_bands(ds):
            b.DeleteNoDataValue()


def get_raster_minimum(filename_or_ds):
    with OpenDS(filename_or_ds) as ds:
        return min(_band_getmin(b) for b in _get_bands(ds))


def get_image_structure_metadata(filename_or_ds, key: str, default=None):
    key = key.strip()
    if not key.endswith("="):
        key = key + "="
    with OpenDS(filename_or_ds) as ds:
        metadata = ds.GetMetadata_List("IMAGE_STRUCTURE")
        if metadata is None:
            return default
        for metadata in metadata:
            if metadata.startswith(key):
                return metadata[len(key) :]
        return default


def expand_txt(filename):
    # input argument is a txt file, replace it with a list of its lines
    filename = Path(filename.strip())
    with open(filename) as f:
        return f.read().splitlines()


def flatten_and_expand_file_list(l, do_expand_txt=True, do_expand_glob=True):
    if is_path_like(l):
        item = str(l).strip()
        if do_expand_glob:
            item1 = glob.glob(str(item).strip())
            if len(item1) == 1:
                item = item1[0]
            elif len(item1) > 1:
                return flatten_and_expand_file_list(item1)

        if (
            do_expand_txt
            and os.path.isfile(item)
            and not os.path.isdir(item)
            and Path(item).suffix.lower() == ".txt"
        ):
            return flatten_and_expand_file_list(expand_txt(item))
        else:
            return item.strip()

    if not is_list_like(l):
        return l
    flat_list = []
    for item in l:
        item1 = flatten_and_expand_file_list(item)
        if is_list_like(item1):
            flat_list.extend(item1)
        else:
            flat_list.append(item1)
    return flat_list


def is_path_like(s):
    return isinstance(s, (str, Path))


def is_list_like(lst):
    return isinstance(lst, Sequence) and not isinstance(lst, str)


def concat_paths(*argv):
    return Path("".join([str(p) for p in argv]))


def wkt_write_ogr(path, wkt_list, of='ESRI Shapefile', epsg=4326):
    driver = ogr.GetDriverByName(of)
    ds = driver.CreateDataSource(path)
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(epsg)

    layer = ds.CreateLayer('', srs, ogr.wkbUnknown)
    for wkt in wkt_list:
        feature = ogr.Feature(layer.GetLayerDefn())
        geom = ogr.CreateGeometryFromWkt(wkt)
        feature.SetGeometry(geom)  # Set the feature geometry
        layer.CreateFeature(feature)  # Create the feature in the layer
        feature.Destroy()  # Destroy the feature to free resources
    # Destroy the data source to free resources
    ds.Destroy()


def get_ext_by_of(of: str):
    ext = of.lower()
    if ext == 'gtiff':
        ext = 'tif'
    return '.'+ext


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ******************************************************************************
#
#  Project:  GDAL
#  Purpose:  Command line raster calculator with numpy syntax
#  Author:   Chris Yesson, chris.yesson@ioz.ac.uk
#
# ******************************************************************************
#  Copyright (c) 2010, Chris Yesson <chris.yesson@ioz.ac.uk>
#  Copyright (c) 2010-2011, Even Rouault <even dot rouault at spatialys.com>
#  Copyright (c) 2016, Piers Titus van der Torren <pierstitus@gmail.com>
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
# ******************************************************************************

################################################################
# Command line raster calculator with numpy syntax. Use any basic arithmetic supported by numpy arrays such as +-*\ along with logical operators such as >.  Note that all files must have the same dimensions, but no projection checking is performed.  Use gdal_calc.py --help for list of options.

# example 1 - add two files together
# gdal_calc.py -A input1.tif -B input2.tif --outfile=result.tif --calc="A+B"

# example 2 - average of two layers
# gdal_calc.py -A input.tif -B input2.tif --outfile=result.tif --calc="(A+B)/2"

# example 3 - set values of zero and below to null
# gdal_calc.py -A input.tif --outfile=result.tif --calc="A*(A>0)" --NoDataValue=0
################################################################

from optparse import OptionParser, OptionConflictError, Values
import os
import os.path
import sys
import shlex
import string
from collections import defaultdict
from pathlib import Path

import numpy

from osgeo import gdal
from osgeo import gdalnumeric
from gdalos import gdalos_extent, gdalos_color

# create alphabetic list (lowercase + uppercase) for storing input layers
AlphaList = list(string.ascii_letters)

# set up some default nodatavalues for each datatype
DefaultNDVLookup = {'Byte': 255, 'UInt16': 65535, 'Int16': -32768, 'UInt32': 4294967293, 'Int32': -2147483647,
                    'Float32': 3.402823466E+38, 'Float64': 1.7976931348623158E+308}


def DoesDriverHandleExtension(drv, ext):
    exts = drv.GetMetadataItem(gdal.DMD_EXTENSIONS)
    return exts is not None and exts.lower().find(ext.lower()) >= 0


def GetExtension(filename):
    ext = os.path.splitext(filename)[1]
    if ext.startswith('.'):
        ext = ext[1:]
    return ext


def GetOutputDriversFor(filename):
    drv_list = []
    ext = GetExtension(filename)
    for i in range(gdal.GetDriverCount()):
        drv = gdal.GetDriver(i)
        if (drv.GetMetadataItem(gdal.DCAP_CREATE) is not None or
            drv.GetMetadataItem(gdal.DCAP_CREATECOPY) is not None) and \
                drv.GetMetadataItem(gdal.DCAP_RASTER) is not None:
            if ext and DoesDriverHandleExtension(drv, ext):
                drv_list.append(drv.ShortName)
            else:
                prefix = drv.GetMetadataItem(gdal.DMD_CONNECTION_PREFIX)
                if prefix is not None and filename.lower().startswith(prefix.lower()):
                    drv_list.append(drv.ShortName)

    # GMT is registered before netCDF for opening reasons, but we want
    # netCDF to be used by default for output.
    if ext.lower() == 'nc' and not drv_list and \
            drv_list[0].upper() == 'GMT' and drv_list[1].upper() == 'NETCDF':
        drv_list = ['NETCDF', 'GMT']

    return drv_list


def GetOutputDriverFor(filename):
    drv_list = GetOutputDriversFor(filename)
    ext = GetExtension(filename)
    if not drv_list:
        if not ext:
            return 'GTiff'
        else:
            raise Exception("Cannot guess driver for %s" % filename)
    elif len(drv_list) > 1:
        print("Several drivers matching %s extension. Using %s" % (ext if ext else '', drv_list[0]))
    return drv_list[0]


################################################################


def doit(opts, args):
    # pylint: disable=unused-argument

    if opts.debug:
        print("gdal_calc.py starting calculation %s" % (opts.calc))

    # set up global namespace for eval with all functions of gdalnumeric
    global_namespace = dict([(key, getattr(gdalnumeric, key))
                             for key in dir(gdalnumeric) if not key.startswith('__')])
    if opts.user_namespace:
        global_namespace.update(opts.user_namespace)

    if not opts.calc:
        raise Exception("No calculation provided.")
    elif not opts.outF:
        if not opts.return_ds:
            raise Exception("No output file provided.")
        else:
            os.format = 'MEM'

    if opts.format is None:
        opts.format = GetOutputDriverFor(opts.outF)

    ################################################################
    # fetch details of input layers
    ################################################################

    # set up some lists to store data for each band
    myFileNames = []
    myFiles = []
    myBands = []
    myAlphaList = []
    myDataType = []
    myDataTypeNum = []
    myNDV = []
    DimensionsCheck = None
    Dimensions = []
    ProjectionCheck = None
    GeoTransformCheck = None
    GeoTransforms = []
    GeoTransformDiffer = False
    myTempFileNames = []
    myFileLists = []

    # loop through input files - checking dimensions
    for myIs, myFs in opts.input_files.items():
        if isinstance(myFs, (list, tuple)):
            # myI is a list of files
            myFileLists.append(myIs)
        elif isinstance(myFs, (str, Path, gdal.Dataset)):
            # myI is a single filename or a Dataset
            myFs = [myFs]
            myIs = [myIs]
        else:
            # I guess myI should be in the global_namespace,
            # It would have been better to pass it as user_namepsace, but I'll accept it anyway
            global_namespace[myIs] = myFs
            continue
        for myI, myF in zip(myIs*len(myFs), myFs):
            if not myI.endswith("_band"):
                # check if we have asked for a specific band...
                if "%s_band" % myI in opts.input_files:
                    myBand = opts.input_files["%s_band" % myI]
                else:
                    myBand = 1

                myF_is_ds = not isinstance(myF, (str, Path))
                if myF_is_ds:
                    myFile = myF
                    myF = None
                else:
                    myF = str(myF)
                    myFile = gdal.Open(myF, gdal.GA_ReadOnly)
                if not myFile:
                    raise IOError("No such file or directory: '%s'" % myF)

                myFileNames.append(myF)
                myFiles.append(myFile)
                myBands.append(myBand)
                myAlphaList.append(myI)
                myDataType.append(gdal.GetDataTypeName(myFile.GetRasterBand(myBand).DataType))
                myDataTypeNum.append(myFile.GetRasterBand(myBand).DataType)
                myNDV.append(None if opts.hideNodata else myFile.GetRasterBand(myBand).GetNoDataValue())

                # check that the dimensions of each layer are the same
                myFileDimensions = [myFile.RasterXSize, myFile.RasterYSize]
                if DimensionsCheck:
                    if DimensionsCheck != myFileDimensions:
                        GeoTransformDiffer = True
                        if opts.extent in [0, 1]:
                            raise Exception("Error! Dimensions of file %s (%i, %i) are different from other files (%i, %i).  Cannot proceed" %
                                            (myF, myFileDimensions[0], myFileDimensions[1], DimensionsCheck[0], DimensionsCheck[1]))
                else:
                    DimensionsCheck = myFileDimensions

                # check that the Projection of each layer are the same
                myProjection = myFile.GetProjection()
                if ProjectionCheck:
                    if opts.projectionCheck and ProjectionCheck != myProjection:
                        raise Exception(
                            "Error! Projection of file %s %s are different from other files %s.  Cannot proceed" %
                            (myF, myProjection, ProjectionCheck))
                else:
                    ProjectionCheck = myProjection

                # check that the GeoTransforms of each layer are the same
                myFileGeoTransform = myFile.GetGeoTransform()
                if opts.extent:
                    Dimensions.append(myFileDimensions)
                    GeoTransforms.append(myFileGeoTransform)
                    if GeoTransformCheck:
                        if GeoTransformCheck != myFileGeoTransform:
                            GeoTransformDiffer = True
                            if opts.extent == 1:
                                raise Exception(
                                    "Error! GeoTransform of file %s %s are different from other files %s.  Cannot proceed" %
                                    (myF, myFileGeoTransform, GeoTransformCheck))
                            else:
                                eps = 0.000001
                                for i in (1, 5):
                                    if abs(GeoTransformCheck[i] - myFileGeoTransform[i])>eps:
                                        raise Exception(
                                            "Error! Pixel size file %s %s are different from other files %s.  Cannot proceed" %
                                            (myF, myFileGeoTransform, GeoTransformCheck))
                                for i in (2, 4):
                                    if abs(GeoTransformCheck[i] - myFileGeoTransform[i])>eps:
                                        raise Exception(
                                            "Error! The rotation of file %s is %s, only 0 is accepted.  Cannot proceed" %
                                            (myF, (myFileGeoTransform[2], myFileGeoTransform[4]), GeoTransformCheck))
                    else:
                        GeoTransformCheck = myFileGeoTransform
                else:
                    GeoTransformCheck = myFileGeoTransform

                if opts.debug:
                    print("file %s: %s, dimensions: %s, %s, type: %s" % (
                    myI, myF, DimensionsCheck[0], DimensionsCheck[1], myDataType[-1]))

    # process allBands option
    allBandsIndex = None
    allBandsCount = 1
    if opts.allBands:
        try:
            allBandsIndex = myAlphaList.index(opts.allBands)
        except ValueError:
            raise Exception("Error! allBands option was given but Band %s not found.  Cannot proceed" % (opts.allBands))
        allBandsCount = myFiles[allBandsIndex].RasterCount
        if allBandsCount <= 1:
            allBandsIndex = None

    new_mode = True
    if opts.extent and (GeoTransformDiffer or not isinstance(opts.extent, int)):
        GeoTransformCheck, DimensionsCheck, ExtentCheck = gdalos_extent.calc_geotransform_and_dimensions(
            GeoTransforms, Dimensions, opts.extent)
        if GeoTransformCheck is None:
            raise Exception("Error! The requested extent is empty. Cannot proceed")
        for i in range(len(myFileNames)):
            if new_mode:
                temp_vrt_filename, temp_vrt_ds = gdalos_extent.make_temp_vrt(myFiles[i], ExtentCheck)
            else:
                temp_vrt_filename, temp_vrt_ds = gdalos_extent.make_temp_vrt_old(
                    myFileNames[i], myFiles[i], myDataType[i], ProjectionCheck, allBandsCount,
                    GeoTransforms[i], Dimensions[i], GeoTransformCheck, DimensionsCheck)
            myTempFileNames.append(temp_vrt_filename)
            myFiles[i] = None  # close original ds
            myFiles[i] = temp_vrt_ds  # replace original ds with vrt_ds

            # update the new precise dimensions and gt from the new ds
            GeoTransformCheck = temp_vrt_ds.GetGeoTransform()
            DimensionsCheck = [temp_vrt_ds.RasterXSize, temp_vrt_ds.RasterYSize]

        temp_vrt_ds = None


    ################################################################
    # set up output file
    ################################################################

    # open output file exists
    if opts.outF and os.path.isfile(opts.outF) and not opts.overwrite:
        if allBandsIndex is not None:
            raise Exception("Error! allBands option was given but Output file exists, must use --overwrite option!")
        if opts.debug:
            print("Output file %s exists - filling in results into file" % (opts.outF))
        myOut = gdal.Open(opts.outF, gdal.GA_Update)
        if myOut is None:
            raise Exception("Error! output file exists but cannot be opened for update. must use --overwrite option!")

        error = None
        if [myOut.RasterXSize, myOut.RasterYSize] != DimensionsCheck:
            error = 'size'
        elif ProjectionCheck and ProjectionCheck != myOut.GetProjection():
            error = 'projection'
        elif GeoTransformCheck and GeoTransformCheck != myOut.GetGeoTransform():
            error = 'geotransform'
        if error:
            raise Exception("Error! Output exists, but is the wrong %s.  Use the --overwrite option to automatically overwrite the existing file" % error)

        myOutB = myOut.GetRasterBand(1)
        myOutNDV = myOutB.GetNoDataValue()
        myOutType = gdal.GetDataTypeName(myOutB.DataType)

    else:
        # remove existing file and regenerate
        if opts.outF:
            if os.path.isfile(opts.outF):
                os.remove(opts.outF)
            # create a new file
            if opts.debug:
                print("Generating output file %s" % (opts.outF))
        else:
            opts.outF = ''

        # find data type to use
        if not opts.type:
            # use the largest type of the input files
            myOutType = gdal.GetDataTypeName(max(myDataTypeNum))
        else:
            myOutType = opts.type

        # create file
        myOutDrv = gdal.GetDriverByName(opts.format)
        myOut = myOutDrv.Create(
            opts.outF, DimensionsCheck[0], DimensionsCheck[1], allBandsCount,
            gdal.GetDataTypeByName(myOutType), opts.creation_options)

        # set output geo info based on first input layer
        if not GeoTransformCheck:
            GeoTransformCheck = myFiles[0].GetGeoTransform()
        myOut.SetGeoTransform(GeoTransformCheck)
        if not ProjectionCheck:
            ProjectionCheck = myFiles[0].GetProjection()
        myOut.SetProjection(ProjectionCheck)

        if opts.NoDataValue is not None:
            myOutNDV = opts.NoDataValue
        else:
            myOutNDV = DefaultNDVLookup[myOutType]

        for i in range(1, allBandsCount + 1):
            myOutB = myOut.GetRasterBand(i)
            myOutB.SetNoDataValue(myOutNDV)
            if opts.color_table:
                # set color table and color interpretation
                if isinstance(opts.color_table, str):
                    opts.color_table = gdalos_color.get_color_table(opts.color_table)  # todo: use gdal function?
                myOutB.SetRasterColorTable(opts.color_table)
                myOutB.SetRasterColorInterpretation(gdal.GCI_PaletteIndex)

            myOutB = None  # write to band

    if opts.debug:
        print("output file: %s, dimensions: %s, %s, type: %s" % (
        opts.outF, myOut.RasterXSize, myOut.RasterYSize, myOutType))

    ################################################################
    # find block size to chop grids into bite-sized chunks
    ################################################################

    # use the block size of the first layer to read efficiently
    myBlockSize = myFiles[0].GetRasterBand(myBands[0]).GetBlockSize()
    # find total x and y blocks to be read
    nXBlocks = (int)((DimensionsCheck[0] + myBlockSize[0] - 1) / myBlockSize[0])
    nYBlocks = (int)((DimensionsCheck[1] + myBlockSize[1] - 1) / myBlockSize[1])
    myBufSize = myBlockSize[0] * myBlockSize[1]

    if opts.debug:
        print("using blocksize %s x %s" % (myBlockSize[0], myBlockSize[1]))

    # variables for displaying progress
    ProgressCt = -1
    ProgressMk = -1
    ProgressEnd = nXBlocks * nYBlocks * allBandsCount

    ################################################################
    # start looping through each band in allBandsCount
    ################################################################

    for bandNo in range(1, allBandsCount + 1):

        ################################################################
        # start looping through blocks of data
        ################################################################

        # store these numbers in variables that may change later
        nXValid = myBlockSize[0]
        nYValid = myBlockSize[1]

        # loop through X-lines
        for X in range(0, nXBlocks):

            # in case the blocks don't fit perfectly
            # change the block size of the final piece
            if X == nXBlocks - 1:
                nXValid = DimensionsCheck[0] - X * myBlockSize[0]

            # find X offset
            myX = X * myBlockSize[0]

            # reset buffer size for start of Y loop
            nYValid = myBlockSize[1]
            myBufSize = nXValid * nYValid

            # loop through Y lines
            for Y in range(0, nYBlocks):
                ProgressCt += 1
                if 10 * ProgressCt / ProgressEnd % 10 != ProgressMk and not opts.quiet:
                    ProgressMk = 10 * ProgressCt / ProgressEnd % 10
                    from sys import version_info
                    if version_info >= (3, 0, 0):
                        exec('print("%d.." % (10*ProgressMk), end=" ")')
                    else:
                        exec('print 10*ProgressMk, "..",')

                # change the block size of the final piece
                if Y == nYBlocks - 1:
                    nYValid = DimensionsCheck[1] - Y * myBlockSize[1]
                    myBufSize = nXValid * nYValid

                # find Y offset
                myY = Y * myBlockSize[1]

                # create empty buffer to mark where nodata occurs
                myNDVs = None

                # make local namespace for calculation
                local_namespace = {}
                val_lists = defaultdict(list)
                # fetch data for each input layer
                for i, Alpha in enumerate(myAlphaList):

                    # populate lettered arrays with values
                    if allBandsIndex is not None and allBandsIndex == i:
                        myBandNo = bandNo
                    else:
                        myBandNo = myBands[i]
                    myval = gdalnumeric.BandReadAsArray(myFiles[i].GetRasterBand(myBandNo),
                                                        xoff=myX, yoff=myY,
                                                        win_xsize=nXValid, win_ysize=nYValid)
                    if myval is None:
                        raise Exception("Cannot read band array from %s" % myF[i])
                    # fill in nodata values
                    if myNDV[i] is not None:
                        # myNDVs is a boolean buffer.
                        # a cell equals to 1 if there is NDV in any of the corresponsing cells in input raster bands.
                        if myNDVs is None:
                            # this is the first band that has NDV set. we initializes myNDVs to a zero buffer
                            # as we didn't see any NDV value yet.
                            myNDVs = numpy.zeros(myBufSize)
                            myNDVs.shape = (nYValid, nXValid)
                        myNDVs = 1 * numpy.logical_or(myNDVs == 1, myval == myNDV[i])

                    # add an array of values for this block to the eval namespace
                    if Alpha in myFileLists:
                        val_lists[Alpha].append(myval)
                    else:
                        local_namespace[Alpha] = myval
                    myval = None

                for lst in myFileLists:
                    local_namespace[lst] = val_lists[lst]
                # try the calculation on the array blocks
                try:
                    myResult = eval(opts.calc, global_namespace, local_namespace)
                except:
                    print("evaluation of calculation %s failed" % (opts.calc))
                    raise

                # Propagate nodata values (set nodata cells to zero
                # then add nodata value to these cells).
                if myNDVs is not None:
                    myResult = ((1 * (myNDVs == 0)) * myResult) + (myOutNDV * myNDVs)
                elif not isinstance(myResult, numpy.ndarray):
                    myResult = numpy.ones((nYValid, nXValid)) * myResult

                # write data block to the output file
                myOutB = myOut.GetRasterBand(bandNo)
                gdalnumeric.BandWriteArray(myOutB, myResult, xoff=myX, yoff=myY)
                myOutB = None  # write to band

    for idx, tempFile in enumerate(myTempFileNames):
        myFiles[idx] = None
        os.remove(tempFile)
    if not opts.quiet:
        print("100 - Done")
    if not opts.return_ds:
        myOut = None  # delete ds if outfile is given. otherwise return ds
    return myOut

################################################################


def Calc(calc, outfile, NoDataValue=None, type=None, format=None, creation_options=None, allBands='', overwrite=False,
         debug=False, quiet=False, hideNodata=False, projectionCheck=False, color_table=None,
         extent=None, return_ds=False, user_namespace=None, **input_files):
    """ Perform raster calculations with numpy syntax.
    Use any basic arithmetic supported by numpy arrays such as +-*\ along with logical
    operators such as >. Note that all files must have the same dimensions, but no projection checking is performed.

    Keyword arguments:
        [A-Z]: input files
        [A_band - Z_band]: band to use for respective input file

    Examples:
    add two files together:
        Calc("A+B", A="input1.tif", B="input2.tif", outfile="result.tif")

    average of two layers:
        Calc(calc="(A+B)/2", A="input1.tif", B="input2.tif", outfile="result.tif")

    set values of zero and below to null:
        Calc(calc="A*(A>0)", A="input.tif", A_band=2, outfile="result.tif", NoDataValue=0)

    --calc="sum(a,axis=0)" --outfile sum.tif --color_table sum.txt --hideNodata --extent=2 --overwrite -a -1_-1.tif -a -1_0.tif -a -1_1.tif -a 0_-1.tif -a 0_0.tif -a 0_1.tif -a 1_-1.tif -a 1_0.tif -a 1_1.tif
    """
    opts = Values()
    opts.input_files = input_files
    opts.calc = calc
    opts.outF = outfile
    opts.NoDataValue = NoDataValue
    opts.type = type
    opts.format = format
    opts.creation_options = [] if creation_options is None else creation_options
    opts.allBands = allBands
    opts.overwrite = overwrite
    opts.debug = debug
    opts.quiet = quiet

    opts.return_ds = return_ds
    opts.hideNodata = hideNodata
    opts.projectionCheck = projectionCheck
    opts.color_table = color_table
    opts.extent = extent
    opts.user_namespace = user_namespace

    return doit(opts, None)


def store_input_file(option, opt_str, value, parser):
    # pylint: disable=unused-argument
    if not hasattr(parser.values, 'input_files'):
        parser.values.input_files = defaultdict(list)
    key = opt_str.lstrip('-')
    parser.values.input_files[key].append(value)


def add_alpha_args(parser, argv):
    # limit the input file options to the ones in the argument list
    given_args = set([a[1] for a in argv if a[1:2] in AlphaList] + ['A'])
    for myAlpha in given_args:
        try:
            parser.add_option("-%s" % myAlpha, action="callback", callback=store_input_file, type=str,
                              help="input gdal raster file, you can use any letter (A-Z)", metavar='filename')
            parser.add_option("--%s_band" % myAlpha, action="callback", callback=store_input_file, type=int,
                              help="number of raster band for file %s (default 1)" % myAlpha, metavar='n')
        except OptionConflictError:
            pass


def main():
    usage = """usage: %prog --calc=expression --outfile=out_filename [-A filename]
                    [--A_band=n] [-B...-Z filename] [other_options]"""
    parser = OptionParser(usage)

    # define options
    parser.add_option("--calc", dest="calc",
                      help="calculation in gdalnumeric syntax using +-/* or any numpy array functions (i.e. log10())",
                      metavar="expression")
    add_alpha_args(parser, sys.argv)

    parser.add_option("--outfile", dest="outF", help="output file to generate or fill", metavar="filename")
    parser.add_option("--NoDataValue", dest="NoDataValue", type=float,
                      help="output nodata value (default datatype specific value)", metavar="value")
    parser.add_option("--hideNodata", dest="hideNodata", action="store_true",
                      help="ignores the NoDataValues of the input rasters", metavar="value")
    parser.add_option("--type", dest="type", help="output datatype, must be one of %s" % list(DefaultNDVLookup.keys()),
                      metavar="datatype")
    parser.add_option("--format", dest="format", help="GDAL format for output file", metavar="gdal_format")
    parser.add_option(
        "--creation-option", "--co", dest="creation_options", default=[], action="append",
        help="Passes a creation option to the output format driver. Multiple "
             "options may be listed. See format specific documentation for legal "
             "creation options for each format.", metavar="option")
    parser.add_option("--allBands", dest="allBands", default="", help="process all bands of given raster (a-z, A-Z)",
                      metavar="[a-z, A-Z]")
    parser.add_option("--overwrite", dest="overwrite", action="store_true",
                      help="overwrite output file if it already exists")
    parser.add_option("--debug", dest="debug", action="store_true", help="print debugging information")
    parser.add_option("--quiet", dest="quiet", action="store_true", help="suppress progress messages")
    parser.add_option("--optfile", dest="optfile", metavar="optfile",
                      help="Read the named file and substitute the contents into the command line options list.")
    parser.add_option("--extent", dest="extent", type=int,
                      help="how to treat different geotrasnforms [0=ignore|1=fail|2=union|3=intersect]")
    # when extent don't agree: 0=ignore(check only dims)/1=fail (gt must also agree)/2=union/3=intersection/GeoRectangle
    parser.add_option("--projectionCheck", dest="projectionCheck", action="store_true",
                      help="check that all rasters share the same projection", metavar="value")
    parser.add_option("--color_table", dest="color_table", help="color table file name", metavar="filename")

    (opts, args) = parser.parse_args()
    opts.return_ds = False
    opts.user_namespace = None

    if not hasattr(opts, "input_files"):
        opts.input_files = {}

    if opts.optfile:
        with open(opts.optfile, 'r') as f:
            ofargv = [x for line in f for x in shlex.split(line, comments=True)]
        # Avoid potential recursion.
        parser.remove_option('--optfile')
        add_alpha_args(parser, ofargv)
        ofopts, ofargs = parser.parse_args(ofargv)
        # Let options given directly override the optfile.
        input_files = getattr(ofopts, 'input_files', {})
        input_files.update(opts.input_files)
        ofopts.__dict__.update({k: v for k, v in vars(opts).items() if v})
        opts = ofopts
        opts.input_files = input_files
        args = args + ofargs

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    elif not opts.calc:
        print("No calculation provided. Nothing to do!")
        parser.print_help()
        sys.exit(1)
    elif not opts.outF:
        print("No output file provided. Cannot proceed.")
        parser.print_help()
        sys.exit(1)
    else:
        try:
            doit(opts, args)
        except IOError as e:
            print(e)
            sys.exit(1)


if __name__ == "__main__":
    main()


from osgeo import ogr
import os
from typing import Sequence


def create_layer_from_geometries(geoms: Sequence[ogr.Geometry], out_filename, is_ring_geom=True, driver_name='gpkg', layer_name='1'):
    # Remove output shapefile if it already exists
    driver = ogr.GetDriverByName(driver_name)
    if os.path.exists(out_filename):
        driver.DeleteDataSource(out_filename)

    # Create the output shapefile
    ds = driver.CreateDataSource(out_filename)
    outLayer = ds.CreateLayer(layer_name, geom_type=ogr.wkbPolygon)

    # Add an ID field
    # idField = ogr.FieldDefn("id", ogr.OFTInteger)
    # outLayer.CreateField(idField)

    # Create the feature and set values
    featureDefn = outLayer.GetLayerDefn()
    for geom in geoms:
        if is_ring_geom:
            ring = geom
            geom = ogr.Geometry(ogr.wkbPolygon)
            geom.AddGeometry(ring)
        feature = ogr.Feature(featureDefn)
        feature.SetGeometry(geom)
        # feature.SetField("id", 1)
        outLayer.CreateFeature(feature)

    # Close DataSource
    # ds.Destroy()
    del ds



from osgeo import ogr
from gdalos.talos.math0 import SinCos
from gdalos.talos.geom_util import GetFromToAngle
from gdalos.talos.gen_consts import M_2PI, M_DEG2RAD
from gdalos.talos.ogr_util import create_layer_from_geometries
import tempfile


def PolygonizeSector(px, py, rx, ry: float, DirectionDeg, ApertureDeg, ThetaDeg: float=0, PointCount: int=50):
    SinTheta, CosTheta = SinCos(ThetaDeg * M_DEG2RAD)
    # Create ring
    ring = ogr.Geometry(ogr.wkbLinearRing)

    if ApertureDeg == 360:
        Factor = M_2PI / PointCount
        for I in range(0, PointCount+1):
            SinT, CosT = SinCos(I * Factor)
            x = px + rx * CosT * CosTheta - ry * SinT * SinTheta
            y = py + ry * SinT * CosTheta + rx * CosT * SinTheta
            ring.AddPoint(x, y)
    else:
        ring.AddPoint(px, py)
        AFromRad, AToRad = GetFromToAngle(DirectionDeg, ApertureDeg)
        Factor = (AToRad - AFromRad) / (PointCount - 2)
        for I in range(0, PointCount - 1):
            SinT, CosT = SinCos(I * Factor + AFromRad)
            x = px + rx * CosT * CosTheta - ry * SinT * SinTheta
            y = py + ry * SinT * CosTheta + rx * CosT * SinTheta
            ring.AddPoint(x, y)
        ring.AddPoint(px, py)  # close the ring

    # Create polygon
    # poly = ogr.Geometry(ogr.wkbPolygon)
    # poly.AddGeometry(ring)
    return ring


if __name__ == '__main__':
    px, py = 680000.00, 3540000.00
    rx = ry = 1000
    DirectionDeg = 20
    ApertureDeg = 30
    ThetaDeg = 0
    PointCount = 50
    ring = PolygonizeSector(px, py, rx, ry, DirectionDeg, ApertureDeg, ThetaDeg, PointCount)
    geom = ogr.Geometry(ogr.wkbPolygon)
    geom.AddGeometry(ring)
    wkt = geom.ExportToIsoWkt()
    print(wkt)

    out_filename = tempfile.mktemp(suffix='.gpkg')
    # create_layer_from_geometries([geom], out_filename, is_ring_geom=False)
    create_layer_from_geometries([ring], out_filename)
    print(out_filename)



import logging


def set_logger_console(logger, level=logging.INFO):
    logger.setLevel(logging.DEBUG)
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(level)

    # create formatter and add it to the handlers
    formatter = logging.Formatter("%(message)s")
    ch.setFormatter(formatter)

    # add the handlers to logger
    logger.addHandler(ch)
    return ch


def set_file_logger(logger, log_filename, level=logging.DEBUG):
    # create file handler which logs even debug messages
    fh = logging.FileHandler(log_filename, mode="w")
    fh.setLevel(level)

    # create formatter and add it to the handlers
    # formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    fh.setFormatter(formatter)

    # add the handlers to logger
    logger.addHandler(fh)
    # logger.info('test')
    return fh


def test_log1():
    logger = logging.getLogger(__name__)
    logger.info("x")
    set_logger_console(logger)
    logger.info("y")
    set_file_logger(logger, "a.txt")
    logger.info("a")
    set_file_logger(logger, "b.txt")
    logger.info("b")
    set_file_logger(logger, "c.txt")
    logger.info("c")


if __name__ == "__main__":
    test_log1()



import os
from pathlib import Path
import re
from gdalos.viewshed.viewshed_grid_params import ViewshedGridParams
import talos


def talos_run(vp: ViewshedGridParams, output_path, func_name, classname, comb_classname, save_to_file):
    talos.DeleteObjects(classname)
    talos.SetScriptWGSGeo(False)

    arr = vp.get_array()

    output_path1 = output_path / func_name
    os.makedirs(output_path1, exist_ok=True)
    output_path2 = output_path / comb_classname
    os.makedirs(output_path2, exist_ok=True)

    for vp1 in arr:
        f = talos.CreateObject(classname)
        talos.SetVal(f, 'Name', vp1.name)
        talos.SetVal(f, 'Geometry', vp1.oxy)
        talos.SetVal(f, 'Range', vp1.max_r)
        talos.SetVal(f, 'Height', vp1.oz)
        talos.SetVal(f, 'TargetHeight', vp1.tz)
        talos.Run(func_name)
        if save_to_file:
            filename = output_path1 / (vp1.name + '.tif')
            # talos.ShowMessage(filename)
            talos.RasterSaveToFile(f, str(filename), False)
    if save_to_file:
        combined = talos.GetObjects(comb_classname)[0]
        filename = output_path2 / 'combined.tif'
        # talos.ShowMessage(filename)
        talos.RasterSaveToFile(combined, str(filename), False)


if __name__ == "__main__":
    output_path = Path(r'd:\dev\gis\maps\comb_ViewshedBackend.TALOS0')
    save_to_file = True
    vp = ViewshedGridParams()
    func_name = 'FieldOfSight'
    classname = 'T'+func_name
    comb_classname = 'TGlobal' + re.sub('[a-z]', '', func_name)

    talos_run(vp, output_path, func_name, classname, comb_classname, save_to_file)



from copy import copy
from typing import Sequence

st_seen = 5
st_seenbut = 4
st_hidbut = 3
st_hidden = 2
st_nodtm = 1
st_nodata = 0  # out of range value

viewshed_visible = st_seen
viewshed_thresh = st_hidbut
viewshed_invisible = st_hidden
viewshed_out_of_range = st_nodata
viewshed_ndv = st_nodata
viewshed_comb_ndv = 255
viewshed_comb_multi_val = 254

atmospheric_refraction_coeff = 1/7


class ViewshedParams(object):
    __slots__ = ['max_r', 'min_r', 'min_r_shave', 'max_r_slant',
                 'ox', 'oy', 'oz', 'tz', 'omsl', 'tmsl',
                 'azimuth', 'h_aperture', 'elevation', 'v_aperture',
                 'vv', 'iv', 'ov', 'ndv',
                 'refraction_coeff', 'mode']

    def __init__(self):
        self.min_r = 0
        self.max_r = None

        self.min_r_shave = False
        self.max_r_slant = True

        self.ox = None
        self.oy = None
        self.oz = None
        self.tz = None
        self.omsl = False  # observer MSL
        self.tmsl = False  # target MSL

        self.azimuth = 0
        self.h_aperture = 360
        self.elevation = 0
        self.v_aperture = 180

        self.vv = viewshed_visible
        self.iv = viewshed_invisible
        self.ov = viewshed_out_of_range
        self.ndv = viewshed_ndv

        self.refraction_coeff = atmospheric_refraction_coeff
        self.mode = 2

    def is_omni_h(self):
        return not self.h_aperture or abs(self.h_aperture - 360) < 0.0001

    @property
    def oxy(self):
        return self.ox, self.oy

    @oxy.setter
    def oxy(self, oxy):
        self.ox, self.oy = oxy

    def get_as_gdal_params(self):
        vp_params = \
            'max_r', 'ox', 'oy', 'oz', 'tz', \
            'vv', 'iv', 'ov', 'ndv', 'mode'

        gdal_params = \
            'maxDistance', 'observerX', 'observerY', 'observerHeight', 'targetHeight', \
            'visibleVal', 'invisibleVal', 'outOfRangeVal', 'noDataVal', 'mode'
        d = {k1: getattr(self, k0) for k0, k1 in
             zip(vp_params, gdal_params)}
        d['dfCurvCoeff'] = 1 - self.refraction_coeff
        return d

    def get_as_talos_params(self):
        vp_params = \
            'ox', 'oy', 'oz', 'max_r', 'min_r', 'min_r_shave', 'max_r_slant', 'tz', \
            'omsl', 'tmsl', 'azimuth', 'h_aperture', 'elevation', 'v_aperture'

        talos_params = \
            'ox', 'oy', 'oz', 'MaxRange', 'MinRange', 'MinRangeShave', 'SlantRange', 'TargetHeight', \
            'ObsMSL', 'TarMSL', 'Direction', 'Aperture', 'Elevation', 'ElevationAperture'
        d = {k1: getattr(self, k0) for k0, k1 in
             zip(vp_params, talos_params)}
        return d

    def update(self, d: dict):
        for k, v in d.items():
            setattr(self, k, v)

    @staticmethod
    def get_list_from_lists_dict(d: dict, key_map=None) -> Sequence['ViewshedParams']:
        max_len = max(len(v) if v else 0 for v in d.values())
        result = []
        vp = ViewshedParams()
        for i in range(max_len):
            vp = copy(vp)
            for k, v in d.items():  # zip(new_keys, d.values()):
                if not v:
                    continue
                if key_map:
                    k = key_map[k]
                len_v = len(v)
                if i < len_v:
                    setattr(vp, k, v[i])
                    # vp1.k = v[i]
            result.append(vp)
        return result


# gdal_viewshed_params_short = \
#     'max_r', 'ox', 'oy', 'oz', 'tz', \
#     'vv', 'iv', 'ov', 'ndv', 'mode',
#
# gdal_viewshed_params_full = \
#     'maxDistance', 'observerX', 'observerY', 'observerHeight', 'targetHeight', \
#     'visibleVal', 'invisibleVal', 'outOfRangeVal', 'noDataVal', 'mode'
# gdal_viewshed_keymap = dict(zip(gdal_viewshed_params_short, gdal_viewshed_params_full))

viewshed_defaults = dict(vv=viewshed_visible,
                         iv=viewshed_invisible,
                         ov=viewshed_out_of_range,
                         ndv=viewshed_ndv,
                         )

# gdal_viewshed_defaults = dict_util.replace_keys(viewshed_defaults,  gdal_viewshed_keymap)



import os
import glob
import gdal
import tempfile
from pathlib import Path
from gdalos import gdalos_util, gdalos_extent, GeoRectangle
from gdalos.calc import gdal_calc
from gdalos.gdalos_color import ColorPalette
from gdalos.calc.gdalcompare import compare
from gdalos.calc import gdalos_combine


def do_comb(filenames, alpha_pattern, operand='+', **kwargs):
    calc, kwargs = gdalos_combine.make_calc_with_operand(filenames, alpha_pattern, operand, **kwargs)
    gdal_calc.Calc(calc, **kwargs)


def make_verts(filenames, isUnion):
    # extents = []
    dss = []
    res_extent = None
    for filename in filenames:
        ds = gdalos_util.open_ds(filename)
        extent = gdalos_extent.get_extent(ds)

        dss.append(ds)
        # extents.append(extent)
        res_extent = extent if res_extent is None else res_extent.union(extent) if isUnion else res_extent.intersect(extent)

    suffix = '_u.vrt' if isUnion else '_i.vrt'
    vrt_filenames = build_vrts(filenames, dss, res_extent, suffix)
    return vrt_filenames, res_extent


# def combine(filenames, outpath, alpha_pattern, **kwargs):
#     extents = []
#     dss = []
#     union_extent = None
#     intersect_extent = None
#     for filename in filenames:
#         ds = gdalos_util.open_ds(filename)
#         extent = gdalos_extent.get_extent(ds)
#
#         dss.append(ds)
#         extents.append(extent)
#         if union_extent is None:
#             union_extent = extent
#             intersect_extent = extent
#         else:
#             intersect_extent = intersect_extent.intersect(extent)
#             union_extent = union_extent.union(extent)
#
#     vrt_filenames = build_vrts(filenames, dss, union_extent, '_u.vrt')
#     outfile = tempfile.mktemp(suffix='_union_combine.tif', dir=str(outpath))
#     # outfile = outpath / suffix
#     do_comb(vrt_filenames, alpha_pattern, outfile=outfile, **kwargs)
#
#     vrt_filenames = build_vrts(filenames, dss, intersect_extent, '_i.vrt')
#     outfile = tempfile.mktemp(suffix='_intersect_combine.tif', dir=str(outpath))
#     # outfile = outpath / suffix
#     do_comb(vrt_filenames, alpha_pattern, outfile=outfile, **kwargs)
#     return extents


def build_vrts(filenames, dss, extent: GeoRectangle, suffix):
    vrt_filenames = []
    for filename, ds in zip(filenames, dss):
        options = gdal.BuildVRTOptions(outputBounds=(extent.min_x, extent.min_y, extent.max_x, extent.max_y))
                                       # hideNodata=True,
                                       # separate=False)
        # vrt_filename = filename + suffix
        vrt_filename = tempfile.mktemp(suffix='.vrt')
        vrt_filenames.append(vrt_filename)
        vrt_ds = gdal.BuildVRT(vrt_filename, ds, options=options)
        if vrt_ds is None:
            return None
        del vrt_ds
    return vrt_filenames


def compare_rasters(pattern):
    golden = None
    total = 0
    for filename in glob.glob(str(pattern)):
        if golden is None:
            golden = filename
        else:
            print('{} vs {}'.format(golden, filename))
            res = compare(golden, filename)
            print(res)
            total += res
    print('total: {}'.format(total))
    return total


def combine_all(path, outpath):
    pattern = Path('*.tif')
    os.makedirs(str(outpath), exist_ok=True)
    dirpath = path / pattern
    filenames = list(glob.glob(str(dirpath)))
    print(filenames)

    custom_extent = GeoRectangle.from_lrud(698386, 700147, 3552332, 3550964)
    for extent in [2, 3, custom_extent]:
        for method in [1, 2]:
            isUnion = extent == 2
            isIntersection = extent == 3
            outfile = tempfile.mktemp(suffix='_{}{}.tif'.format(
                'u' if isUnion else 'i' if isIntersection else 'c', method), dir=str(outpath))
            kwargs = dict()
            if method == 1:
                kwargs['filenames'] = filenames
                kwargs['extent'] = extent
            else:
                if not isUnion and not isIntersection:
                    continue
                vrt_filenames, c_extent = make_verts(filenames, isUnion)
                kwargs['filenames'] = vrt_filenames

            do_comb(alpha_pattern=alpha_pattern, outfile=outfile, color_table=color_table, hideNodata=True, **kwargs)


if __name__ == '__main__':
    path = Path(r'sample')
    color_filename = path / Path(r'color_files/viewshed/sum.qlr')
    pal = ColorPalette()
    pal.read(color_filename)
    pal.write_color_file(color_filename.with_suffix('.txt'))
    color_table = pal.get_color_table()
    alpha_pattern = '1*({}>3)'

    path = Path('/home/idan/maps/comb/vs')
    outpath = path / 'comb'

    combine_all(path, outpath)
    total = compare_rasters(outpath / '*_u*') + \
        compare_rasters(outpath / '*_i*') + \
        compare_rasters(outpath / '*_c*')
    print('grand total: {}'.format(total))



from enum import Enum
import sys
import gdal, osr, ogr
import glob
import tempfile
import os
import time
from pathlib import Path
from enum import Enum

from attr import exceptions

from gdalos import projdef, gdalos_util, gdalos_color, gdalos_trans
from gdalos.calc import gdal_calc, gdal_to_czml, dict_util, gdalos_combine
from gdalos.viewshed import viewshed_params
from gdalos.viewshed.viewshed_params import ViewshedParams
from gdalos.viewshed.viewshed_grid_params import ViewshedGridParams
from gdalos.talos.ogr_util import create_layer_from_geometries
from gdalos.talos.geom_arc import PolygonizeSector

talos = None


class ViewshedBackend(Enum):
    gdal = 0
    talos = 1


default_ViewshedBackend = ViewshedBackend.gdal


class CalcOperation(Enum):
    viewshed = 0
    max = 1
    count = 2
    count_z = 3
    unique = 4


def tempthing(use_temp_tif, steps, output_filename):
    is_temp_file = (use_temp_tif and steps > 1)
    gdal_out_format = 'GTiff' if steps == 1 or use_temp_tif else 'MEM'
    d_path = tempfile.mktemp(
        suffix='.tif') if is_temp_file else output_filename if gdal_out_format != 'MEM' else ''
    return_ds = not is_temp_file
    return is_temp_file, gdal_out_format, d_path, return_ds


def make_slice(slicer):
    if isinstance(slicer, slice):
        return slicer
    if slicer is None:
        return slice(None)
    return slice(*[{True: lambda n: None, False: int}[x == ''](x) for x in (slicer.split(':') + ['', '', ''])[:3]])


def viewshed_calc(vp_array,
                  output_filename,
                  input_ds,
                  input_filename=None,
                  extent=2, cutline=None, operation: CalcOperation = CalcOperation.count,
                  in_coords_crs_pj=None, out_crs=None,
                  color_palette=None,
                  bi=1, co=None, of='GTiff',
                  vp_slice=None,
                  backend:ViewshedBackend=None,
                  files=[]):
    if output_filename:
        os.makedirs(os.path.dirname(str(output_filename)), exist_ok=True)
    ext = gdalos_util.get_ext_by_of(of)
    is_czml = ext == '.czml'
    color_table = gdalos_color.get_color_table(color_palette)
    # steps:
    # 1. viewshed
    # 2. calc + calc_post_process
    # 3. combined_post_process
    # 4. czml
    steps = 1
    if operation == CalcOperation.viewshed:
        operation = None
    if operation:
        steps += 1
    if is_czml:
        out_crs = 0  # czml supprts only 4326
        steps += 1
    temp_files = []

    combined_post_process_needed = False

    vp_slice = make_slice(vp_slice)

    if not files:
        files = []
    else:
        files = files.copy()[vp_slice]

    if input_filename is not None:
        input_filename = Path(input_filename).resolve()
        if input_ds is None:
            input_ds = gdalos_util.open_ds(input_filename)
    if input_ds is None:
        if not files:
            raise Exception('ds is None')
    else:
        input_band: gdal.Band = input_ds.GetRasterBand(bi)
        if input_band is None:
            raise Exception('band number out of range')

        pjstr_src_srs = projdef.get_srs_pj_from_ds(input_ds)
        pjstr_tgt_srs = projdef.get_proj_string(out_crs) if out_crs is not None else pjstr_src_srs
        combined_post_process_needed = cutline or not projdef.proj_is_equivalent(pjstr_src_srs, pjstr_tgt_srs)
        if combined_post_process_needed:
            steps += 1

    if not files:
        if isinstance(vp_array, ViewshedParams):
            vp_array = [vp_array]
        else:
            if isinstance(vp_array, dict):
                vp_array = ViewshedGridParams.get_list_from_lists_dict(vp_array)
            vp_array = vp_array[vp_slice]

        if operation:
            # restore viewshed consts default values
            my_viewshed_defaults = viewshed_params.viewshed_defaults
            for a in vp_array:
                a.update(my_viewshed_defaults)
        else:
            vp_array = vp_array[0:1]

        max_rasters_count = 1 if operation is None else 254 if operation == CalcOperation.unique else 1000
        if len(vp_array) > max_rasters_count:
            vp_array = vp_array[0:max_rasters_count]

        # in_raster_srs = projdef.get_srs_pj_from_ds(input_ds)
        in_raster_srs = osr.SpatialReference()
        in_raster_srs.ImportFromWkt(input_ds.GetProjection())
        if not in_raster_srs.IsProjected:
            raise Exception(f'input raster has to be projected')

        if in_coords_crs_pj is not None:
            in_coords_crs_pj = projdef.get_proj_string(in_coords_crs_pj)
            transform_coords_to_raster = projdef.get_transform(in_coords_crs_pj, in_raster_srs)
        else:
            transform_coords_to_raster = None

        for vp in vp_array:
            if transform_coords_to_raster:
                vp.ox, vp.oy, _ = transform_coords_to_raster.TransformPoint(vp.ox, vp.oy)

            if backend is None:
                backend = default_ViewshedBackend
            elif isinstance(backend, str):
                backend = ViewshedBackend[backend]

            inter_process = (backend == ViewshedBackend.gdal) and not vp.is_omni_h()
            if inter_process:
                steps += 1

            # TypeError: '>' not supported between instances of 'NoneType' and 'int'
            # todo: why dosn't it work without it?
            is_temp_file, gdal_out_format, d_path, return_ds = tempthing(True, steps, output_filename)

            if backend == ViewshedBackend.gdal:
                inputs = vp.get_as_gdal_params()
                ds = gdal.ViewshedGenerate(input_band, gdal_out_format, str(d_path), co, **inputs)
                if not ds:
                    raise Exception('Viewshed calculation failed')
            elif backend == ViewshedBackend.talos:
                # is_temp_file = True  # output is file, not ds
                if not input_filename:
                    raise Exception('to use talos backend you need to provide an input filename')
                # if 'talosgis.talos' not in sys.modules:
                global talos
                if talos is None:
                    try:
                        from talosgis import get_talos_gdal_path
                        from talosgis import talos2 as talos
                    except ImportError:
                        raise Exception('failed to load talos backend')

                    print('GS_GetIntVersion ', talos.GS_GetIntVersion())
                    print('GS_GetDLLVersion ', talos.GS_GetDLLVersion())

                    gdal_path = get_talos_gdal_path()
                    # gdal_path = r'd:\OSGeo4W64-20200613\bin\gdal204.dll'
                    talos.GS_SetGDALPath(gdal_path)
                    # print('GS_GetGDALPath ', talos.GS_GetGDALPath())
                    print('GS_talosInit ', talos.GS_talosInit())

                    # print('GS_IsGDALLoaded ', talos.GS_IsGDALLoaded())
                    # print('GS_GetGDALPath ', talos.GS_GetGDALPath())

                    # print('GS_SetCacheSize ', talos.GS_SetCacheSize(cache_size_mb))

                dtm_open_err = talos.GS_DtmOpenDTM(str(input_filename))
                if dtm_open_err != 0:
                    raise Exception('talos could not open input file {}'.format(input_filename))
                talos.GS_SetRefractionCoeff(vp.refraction_coeff)
                inputs = vp.get_as_talos_params()
                ras = talos.GS_Viewshed_Calc1(**inputs)
                talos.GS_SaveRaster(ras, str(d_path))
                ras = None
                # I will reopen the ds to change the color table and ndv
                # ds = gdalos_util.open_ds(d_path, access_mode=gdal.OF_UPDATE)
                ds = gdal.OpenEx(str(d_path), gdal.OF_RASTER | gdal.OF_UPDATE)
            else:
                raise Exception('unknown backend {}'.format(backend))

            src_band = ds.GetRasterBand(1)
            src_band.SetNoDataValue(vp.ndv)
            if color_table and not operation:
                src_band.SetRasterColorTable(color_table)
                src_band.SetRasterColorInterpretation(gdal.GCI_PaletteIndex)
            src_band = None

            if is_temp_file:
                # close original ds and reopen
                ds = None
                ds = gdalos_util.open_ds(d_path)
                temp_files.append(d_path)

            if inter_process:
                ring = PolygonizeSector(vp.ox, vp.oy, vp.max_r, vp.max_r, vp.azimuth, vp.h_aperture)
                calc_cutline = tempfile.mktemp(suffix='.gpkg')
                temp_files.append(calc_cutline)
                create_layer_from_geometries([ring], calc_cutline)

                is_temp_file, gdal_out_format, d_path, return_ds = tempthing(True, steps, output_filename)

                ds = gdalos_trans(ds, out_filename=d_path, #warp_CRS=pjstr_tgt_srs,
                                  cutline=calc_cutline, of=gdal_out_format, return_ds=return_ds, ovr_type=None)

                if is_temp_file:
                    # close original ds and reopen
                    ds = None
                    ds = gdalos_util.open_ds(d_path)
                    temp_files.append(d_path)
                if not ds:
                    raise Exception('Viewshed calculation failed to cut')

                steps -= 1

            if operation:
                files.append(ds)

        input_ds = None
        input_band = None

    steps -= 1
    if operation:
        # alpha_pattern = '1*({{}}>{})'.format(viewshed_thresh)
        # alpha_pattern = 'np.multiply({{}}>{}, dtype=np.uint8)'.format(viewshed_thresh)
        if operation == CalcOperation.viewshed:
            no_data_value = viewshed_params.viewshed_ndv
            f = gdalos_combine.get_by_index
            # calc_expr, calc_kwargs, f = gdal_calc.make_calc_with_func(files, alpha_pattern, 'f'), sum
        elif operation == CalcOperation.max:
            no_data_value = viewshed_params.viewshed_ndv
            f = gdalos_combine.vs_max
            # calc_expr, calc_kwargs, f = gdal_calc.make_calc_with_func(files, alpha_pattern, 'f'), sum
        elif operation == CalcOperation.count:
            no_data_value = 0
            f = gdalos_combine.vs_count
            # calc_expr, calc_kwargs = gdal_calc.make_calc_with_operand(files, alpha_pattern, '+')
            # calc_expr, calc_kwargs, f = gdal_calc.make_calc_with_func(files, alpha_pattern), sum
        elif operation == CalcOperation.count_z:
            no_data_value = viewshed_params.viewshed_comb_ndv
            f = gdalos_combine.vs_count_z
            # calc_expr, calc_kwargs f, = gdal_calc.make_calc_with_func(files, alpha_pattern, 'f'), sum
        elif operation == CalcOperation.unique:
            no_data_value = viewshed_params.viewshed_comb_ndv
            f = gdalos_combine.vs_unique
            # calc_expr, calc_kwargs, f = gdal_calc.make_calc_with_func(files, alpha_pattern, 'f'), unique
        else:
            raise Exception('Unknown operation: {}'.format(operation))

        calc_expr = 'f(x)'
        calc_kwargs = dict(x=files)
        user_namespace = dict(f=f)

        hide_nodata = True

        is_temp_file, gdal_out_format, d_path, return_ds = tempthing(False, steps, output_filename)

        debug_time = 1
        t = time.time()
        for i in range(debug_time):
            ds = gdal_calc.Calc(
                calc_expr, outfile=str(d_path), extent=extent, format=gdal_out_format,
                color_table=color_table, hideNodata=hide_nodata, return_ds=return_ds, overwrite=True,
                NoDataValue=no_data_value, user_namespace=user_namespace, **calc_kwargs)
        t = time.time() - t
        print('time for calc: {:.3f} seconds'.format(t))

        if return_ds:
            if not ds:
                raise Exception('error occurred')
        elif steps > 1:
            ds = gdalos_util.open_ds(d_path)
        for i in range(len(files)):
            files[i] = None  # close calc input ds(s)
        steps -= 1

    if combined_post_process_needed:
        is_temp_file, gdal_out_format, d_path, return_ds = tempthing(False, steps, output_filename)
        ds = gdalos_trans(ds, out_filename=d_path, warp_CRS=pjstr_tgt_srs,
                          cutline=cutline, of=gdal_out_format, return_ds=return_ds, ovr_type=None)

        if return_ds:
            if not ds:
                raise Exception('error occurred')
        elif steps>1:
            ds = gdalos_util.open_ds(d_path)

        steps -= 1

    if is_czml and ds is not None:
        gdal_to_czml.gdal_to_czml(ds, name=output_filename, out_filename=output_filename)

    ds = None  # close ds

    if temp_files:
        for f in temp_files:
            try:
                os.remove(f)
            except:
                print('failed to remove temp file:{}'.format(f))

    return True


if __name__ == "__main__":\
    # dir_path = Path('/home/idan/maps')
    dir_path = Path(r'd:\dev\gis\maps')
    raster_filename = Path(dir_path) / Path('srtm1_36_sample.tif')
    input_ds = ds = gdalos_util.open_ds(raster_filename)

    vp = ViewshedGridParams()

    # inputs = vp.get_as_gdal_params_array()
    inputs = vp.get_array()

    use_input_files = False
    run_comb_with_post = False

    if use_input_files:
        files_path = Path('/home/idan/maps/grid_comb/viewshed')
        files = glob.glob(str(files_path / '*.tif'))
    else:
        files = []

    for backend in reversed(ViewshedBackend):
        # backend = ViewshedBackend.TALOS
        output_path = dir_path / Path('comb_' + str(backend))
        cwd = Path.cwd()
        for calc in CalcOperation:
            # if calc != CalcOperation.viewshed:
            #     continue
            color_palette = cwd / 'sample/color_files/viewshed/{}.txt'.format(calc.name)
            if calc == CalcOperation.viewshed:
                for i, vp in enumerate(inputs):
                    output_filename = output_path / Path('{}_{}.tif'.format(calc.name, i))
                    viewshed_calc(input_ds=input_ds, input_filename=raster_filename,
                                  output_filename=output_filename,
                                  vp_array=vp,
                                  backend=backend,
                                  operation=calc,
                                  color_palette=color_palette,
                                  files=files,
                                  )
            else:
                output_filename = output_path / Path('{}.tif'.format(calc.name))
                viewshed_calc(input_ds=input_ds, input_filename=raster_filename,
                              output_filename=output_filename,
                              vp_array=inputs,
                              backend=backend,
                              operation=calc,
                              color_palette=color_palette,
                              files=files,
                              # vp_slice=slice(0, 2)
                              )

        if run_comb_with_post:
            output_filename = output_path / 'combine_post.tif'
            cutline = cwd / r'sample/shp/comb_poly.gml'
            viewshed_calc(input_ds=input_ds, input_filename=raster_filename,
                          output_filename=output_filename,
                          vp_array=inputs,
                          backend=backend,
                          operation=CalcOperation.count,
                          color_palette=color_palette,
                          cutline=cutline,
                          out_crs=0,
                          files=files)
